#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import contextvars
import logging
import time
import typing
from dataclasses import dataclass, field
from enum import StrEnum
from functools import cached_property
from types import MappingProxyType
from typing import Callable, Any, Iterable, TypeVar, Awaitable, TypeAlias, Literal, ClassVar

import aiohttp
from async_tools import acall
from authlib.jose import JsonWebKey, JsonWebToken, JWTClaims
from authlib.jose.errors import JoseError
from http_tools import IncomingRequest
from init_helpers import raise_if

from .extras import DataclassProtocol, url_to_snake_case
from .parameter import ParameterLocation, QueryParameter, PathParameter, HeaderParameter, SpecParameter

logger = logging.getLogger(__name__)

AuthToken: TypeAlias = str
AuthInfo = TypeVar("AuthInfo", bound=dict | DataclassProtocol)

resolver_cache = contextvars.ContextVar('resolver_cache')


class Unauthorized(Exception):
    pass


class Forbidden(Exception):
    pass


class AuthInfoException(Exception):
    pass


class SecuritySchemeType(StrEnum):
    api_key = "apiKey"
    http = "http"
    oauth2 = "oauth2"
    open_id_connect = "openIdConnect"
    mutual_tls = "mutualTLS"


@dataclass(frozen=True)
class SecurityScheme(abc.ABC, typing.Generic[AuthInfo]):
    type_: SecuritySchemeType
    resolver: Callable[[AuthToken], Awaitable[tuple[AuthInfo, Iterable[str]]] | tuple[AuthInfo, Iterable[str]]]
    do_log: bool = field(default=True, kw_only=True)

    @abc.abstractmethod
    def to_spec(self) -> dict:
        return {'type': self.type_}

    @abc.abstractmethod
    async def _extract_token(self, incoming_request: IncomingRequest) -> AuthToken:
        pass

    async def evaluate(self, incoming_request: IncomingRequest, required_scopes: frozenset[str]) -> AuthInfo:
        cache = resolver_cache.get({})
        if self not in cache:
            auth_token = await self._extract_token(incoming_request)
            try:
                cache[self] = await acall(self.resolver(auth_token))
            except Exception as e:
                cache[self] = e
        result = cache[self]
        if isinstance(result, Exception):
            raise result
        auth_info, allowed_scopes = cache[self]
        allowed_scopes = allowed_scopes if isinstance(allowed_scopes, set) else set(allowed_scopes)
        raise_if(missing_scopes := required_scopes - allowed_scopes,
                 Forbidden(f'missing scopes: {", ".join(sorted(missing_scopes))}'))
        return auth_info

    @property
    @abc.abstractmethod
    def key(self) -> str:
        pass

    def has(self, *args: str, **kwargs: type[Literal] | Callable[[AuthInfo], Any]) -> 'Security':
        return Security(self, scopes=args).use(**kwargs)


@dataclass(frozen=True)
class ApiKeySecurityScheme(SecurityScheme, typing.Generic[AuthInfo]):
    type_: SecuritySchemeType = field(init=False, default_factory=lambda: SecuritySchemeType.api_key)
    location: ParameterLocation
    name: str

    def __post_init__(self):
        assert self._parameter

    def to_spec(self) -> dict:
        return SecurityScheme.to_spec(self) | {'in': self.location, 'name': self.name}

    @cached_property
    def key(self) -> str:
        return f'{self.type_.name}_in_{self.location.name}_{self.name}'

    @cached_property
    def _parameter(self) -> SpecParameter:
        if self.location == ParameterLocation.query:
            return QueryParameter(name=self.name, schema=str)
        if self.location == ParameterLocation.path:
            return PathParameter(name=self.name, schema=str)
        if self.location == ParameterLocation.header:
            return HeaderParameter(name=self.name, schema=str)
        # elif self.in_ == ParameterLocation.cookie: TODO: implement
        #     return CookieParameter(name=self.name, schema=str)
        raise NotImplementedError(f"ApiKeySecurityScheme does not support location {self.location!r} ")

    async def _extract_token(self, incoming_request: IncomingRequest) -> AuthToken:
        try:
            return await self._parameter.get(incoming_request, {})
        except KeyError:
            raise Unauthorized from None


@dataclass(frozen=True)
class HttpSecurityScheme(SecurityScheme, abc.ABC, typing.Generic[AuthInfo]):
    type_: SecuritySchemeType = field(init=False, default_factory=lambda: SecuritySchemeType.http)
    header_scheme: str

    def to_spec(self) -> dict:
        return SecurityScheme.to_spec(self) | {'scheme': self.header_scheme}

    @cached_property
    def key(self) -> str:
        return f'{self.type_.name}_{self.header_scheme}'.strip()

    async def _extract_token(self, incoming_request: IncomingRequest) -> AuthToken:
        dirty_token: str = incoming_request.metadata.header_name_to_value.get("Authorization")
        raise_if(not isinstance(dirty_token, str), Unauthorized("Failed to get token from Authorization header"))
        raise_if(not dirty_token.startswith(self.header_scheme), Unauthorized("Wrong authorization scheme"))
        return dirty_token.removeprefix(self.header_scheme)


@dataclass(frozen=True)
class HttpBearerSecurityScheme(HttpSecurityScheme, typing.Generic[AuthInfo]):
    header_scheme: str = field(init=False, default='Bearer ')


@dataclass(frozen=True)
class HttpBasicSecurityScheme(HttpSecurityScheme, typing.Generic[AuthInfo]):
    header_scheme: str = field(init=False, default='Basic ')


class OAuth2Flow(StrEnum):
    implicit = "implicit"
    authorizationCode = "authorizationCode"
    password = "password"
    clientCredentials = "clientCredentials"


@dataclass(frozen=True)
class OAuth2SecurityScheme(SecurityScheme, abc.ABC, typing.Generic[AuthInfo]):
    type_: SecuritySchemeType = field(init=False, default_factory=lambda: SecuritySchemeType.oauth2)
    # flow_to_info: str ???

    def to_spec(self) -> dict:
        # return SecurityScheme.to_spec(self) | {'in': self.location, 'name': self.name}
        ...


@dataclass(frozen=True, kw_only=True)
class OpenIdConnectSecurityScheme(SecurityScheme, typing.Generic[AuthInfo]):
    url: str
    type_: SecuritySchemeType = field(init=False, default_factory=lambda: SecuritySchemeType.open_id_connect)
    resolver: Callable[[AuthToken], Awaitable[tuple[AuthInfo, Iterable[str]]] | tuple[AuthInfo, Iterable[str]]] = (
        field(init=False))
    header_scheme: ClassVar[str] = 'Bearer'
    expected_issuer: str = field(init=False)
    jwk_lifetime: ClassVar[float] = 60

    def __post_init__(self):
        well_known_index = self.url.find('/.well-known')
        raise_if(well_known_index == -1, ValueError('OpenIdConnectSecurityScheme URL requires ".well-known" part'))
        expected_issuer = self.url[:well_known_index]
        object.__setattr__(self, 'expected_issuer', expected_issuer)

        async def resolver(jwt_value: str) -> tuple[AuthInfo, frozenset[str]]:
            jwks = await self.get_jwks(time.time() - self.jwk_lifetime)
            jwt = JsonWebToken(['RS256'])
            try:
                claims: JWTClaims = jwt.decode(jwt_value, jwks)
                self._check_claims(claims)
            except (JoseError, ValueError) as e:
                raise Unauthorized(str(e)) from None

            return claims, self._get_scopes(claims)

        object.__setattr__(self, 'resolver', resolver)

    @staticmethod
    def _get_scopes(claims: JWTClaims) -> frozenset[str]:
        if (scope := claims.get('scope')) is None:
            raise Unauthorized(f'BadJWT: Missing scope')
        if not isinstance(scope, str):
            raise Unauthorized(f'BadJWT: Scope is not string: {type(scope).__name__}')
        scopes = frozenset(scope.split(' '))
        if 'openid' not in scopes:
            raise Unauthorized(f'BadJWT: Scope MUST contain openid, got: {scope!r}')

        return scopes

    def _check_claims(self, claims: JWTClaims) -> None:
        claims.validate()
        if (issuer := claims.get('iss')) != self.expected_issuer:
            raise Unauthorized(f'BadJWT: Invalid {issuer=}, expected: {self.expected_issuer!r}')
        # TODO: add "aud" check
        # if 'your-client-id' not in (audience := claims.get('aud', [])):
        #     raise Unauthorized(f'BadJWT: Invalid {audience=}')

    async def _extract_token(self, incoming_request: IncomingRequest) -> AuthToken:
        dirty_token: str = incoming_request.metadata.header_name_to_value.get("Authorization")
        raise_if(not isinstance(dirty_token, str), Unauthorized("Failed to get token from Authorization header"))
        raise_if(not dirty_token.startswith(self.header_scheme), Unauthorized("Wrong authorization scheme"))
        return dirty_token.removeprefix(self.header_scheme).strip()

    @property
    def key(self) -> str:
        return f'openid__{url_to_snake_case(self.expected_issuer)}'

    def to_spec(self) -> dict:
        return SecurityScheme.to_spec(self) | {'openIdConnectUrl': self.url}

    async def get_public_keys_url(self) -> str:
        if not hasattr(self, 'public_keys_url'):
            logger.debug(f'get_public_keys_url from {self.url}')
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.get(self.url, timeout=1) as resp:
                        jwks = await resp.json()
                object.__setattr__(self, 'public_keys_url', jwks['jwks_uri'])
            except Exception as e:
                raise RuntimeError("Failed to get public keys url") from e

        return getattr(self, 'public_keys_url')

    async def _load_public_keys(self) -> list[dict]:
        async with aiohttp.ClientSession() as session:
            logger.debug(f'get_public_keys_url from {await self.get_public_keys_url()}')
            async with session.get(await self.get_public_keys_url(), timeout=1) as resp:
                jwks = await resp.json()
        return jwks['keys']

    async def _refresh_public_keys(self) -> None:
        logger.debug('_refresh_public_keys')
        object.__setattr__(self, 'jwks', JsonWebKey.import_key_set({'keys': await self._load_public_keys()}))
        object.__setattr__(self, 'jwks_got_at', time.time())

    async def get_jwks(self, keys_newer_than: float):
        try:
            if getattr(self, 'jwks_got_at') > keys_newer_than:
                return getattr(self, 'jwks')
        except AttributeError:
            pass
        await self._refresh_public_keys()
        return getattr(self, 'jwks')


@dataclass(frozen=True, slots=True)
class Security:
    scheme: SecurityScheme
    scopes: frozenset[str] = field(default_factory=frozenset)
    argument_name_to_getter: MappingProxyType[str, Callable[[AuthInfo], Awaitable | Any]] = field(
        default_factory=dict, hash=False)

    def __init__(self, scheme: SecurityScheme, scopes: Iterable[str] = tuple(),
                 argument_name_to_getter: dict[str, Callable[[AuthInfo], Awaitable | Any]] | None = None):
        object.__setattr__(self, 'scheme', scheme)
        object.__setattr__(self, 'scopes', frozenset(scopes))
        object.__setattr__(self, 'argument_name_to_getter', MappingProxyType(argument_name_to_getter or {}))

    async def evaluate(self, incoming_request: IncomingRequest) -> dict:
        auth_info = await self.scheme.evaluate(incoming_request, self.scopes)
        try:
            return {name: await acall(getter(auth_info)) for name, getter in self.argument_name_to_getter.items()}
        except (KeyError, TypeError, ValueError) as e:
            raise AuthInfoException(*e.args) from e

    def use(self, *args: str, **kwargs: type[Literal] | Callable[[AuthInfo], Any]) -> 'Security':
        return self.has(*args, **kwargs)

    def has(self, *args: str, **kwargs: type[Literal] | Callable[[AuthInfo], Any]) -> 'Security':
        argument_name_to_getter = self.argument_name_to_getter.copy()
        for key, value in kwargs.items():
            if typing.get_origin(value) is Literal:
                args_args = typing.get_args(value)
                raise_if(len(args_args) != 1, TypeError(f'Only single literal values allowed, got: {args_args!r}'))
                argument_name_to_getter[key] = lambda _: args_args[0]
            elif callable(value):
                argument_name_to_getter[key] = value
            else:
                raise TypeError(f'Got unexpected value: {value}, Allowed: Literal/Callable[[AuthInfo], Any]')
        return Security(scheme=self.scheme, scopes=self.scopes | set(args),
                        argument_name_to_getter=argument_name_to_getter)


def get_securities_arg_names(securities: list[Security | SecurityScheme]) -> set[str]:
    result = None
    for security in securities:
        names = set(security.argument_name_to_getter.keys()) if isinstance(security, Security) else set()
        if result is None:
            result = names
        else:
            raise_if(names != result, ValueError(f'All {securities=} must provide same arguments: {names} != {result}'))
    return result or set()



#  Copyright (C) 2024
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import asyncio
import itertools
import operator
import typing
from dataclasses import dataclass, field
from enum import StrEnum
from types import UnionType
from typing import Any, ClassVar, Callable, Annotated, Iterable

from async_tools import acall
from dict_caster.extras import is_iterable
from dynamic_types.class_name import prepare_class_name
from dynamic_types.create import create_type
from http_tools import IncomingRequest
from http_tools.mime_types import ContentType
from http_tools.multipart_form_data import BodyPart
from init_helpers import raise_if, append_if
from init_helpers.dict_to_dataclass import NoValue, convert_to_type, FieldErrors
from multidict import CIMultiDictProxy

from .example import BaseExample


class ParameterLocation(StrEnum):
    query = 'query'
    path = 'path'
    header = 'header'
    cookie = 'cookie'
    body = 'body'  # INNER: not present in specification, used inside library
    security = 'security'  # INNER: not present in specification, used inside library
    literal = 'literal'  # INNER: not present in specification, used inside library


class ParameterStyle(StrEnum):
    """
    Example with values
    string = "blue"
    array  = ["blue","black","brown"]
    object = {"R": 100, "G": 200, "B": 150}
    """

    matrix = 'matrix'                   
    # explode=False     ;color      ;color=blue     ;color=blue,black,brown                 ;color=R,100,G,200,B,150
    # explode=True      ;color      ;color=blue     ;color=blue;color=black;color=brown     ;R=100;G=200;B=150
    label = 'label'
    # explode=False     .           .blue           .blue.black.brown                       .R.100.G.200.B.150
    # explode=True      .           .blue           .blue.black.brown                       .R=100.G=200.B=150
    form = 'form'                       
    # explode=False     color=      color=blue      color=blue,black,brown                  color=R,100,G,200,B,150
    # explode=True      color=      color=blue      color=blue&color=black&color=brown      R=100&G=200&B=150
    simple = 'simple'                   
    # explode=False     n/a         blue            blue,black,brown                        R,100,G,200,B,150
    # explode=True      n/a         blue            blue,black,brown                        R=100,G=200,B=150
    spaceDelimited = 'spaceDelimited'   
    # explode=False     n/a         n/a   	        blue%20black%20brown                    R%20100%20G%20200%20B%20150
    pipeDelimited = 'pipeDelimited'  
    # explode=False     n/a         n/a             blue|black|brown                        R|100|G|200|B|150
    deepObject = 'deepObject'  
    # explode=True      n/a         n/a             n/a                         color[R]=100&color[G]=200&color[B]=150


@dataclass(frozen=True, kw_only=True)
class CallParameter(abc.ABC):
    """Параметр вызова процедуры, schema - тип, key - ключ при вызове"""
    schema: type | UnionType
    # key: str

    @abc.abstractmethod
    async def get(self, incoming_request: IncomingRequest, security_kwargs: dict[str, Any]) -> Any:
        pass

    @abc.abstractmethod
    def get_spec_parameters(self) -> set['SpecParameter']:
        pass


@dataclass(frozen=True, kw_only=True)
class LiteralParameter(CallParameter):
    value: Any

    async def get(self, incoming_request: IncomingRequest, security_kwargs: dict[str, Any]) -> Any:
        return self.value

    def get_spec_parameters(self) -> set['SpecParameter']:
        return set()


@dataclass(frozen=True, kw_only=True)
class SpecParameter(CallParameter, abc.ABC):
    """Параметр, входящий в спецификацию, location - расположение(хедер/query/path/т.п), name - имя в расположении"""
    location: ParameterLocation
    name: str
    is_optional: bool = False
    default: Any = NoValue
    description: str | NoValue = NoValue
    examples: tuple[BaseExample, ...] | NoValue = NoValue
    explode: bool | None = None
    style: ParameterStyle | None = None

    @property
    def annotated_schema(self) -> type | UnionType | Annotated:
        notes = []
        append_if(self.description is not NoValue, notes, self.description)
        append_if(self.examples is not NoValue, notes, *self.examples if self.examples is not NoValue else [])
        return Annotated[self.schema, *notes] if notes else self.schema

    @property
    def is_required(self) -> bool:
        return self.default is NoValue and not self.is_optional

    def get_spec_parameters(self) -> set['SpecParameter']:
        return {self}

    def _cast_value_to_schema(self, value: Any):
        if self.explode:
            type_args = typing.get_args(self.schema)
            if len(type_args) != 1:
                raise TypeError(f'Only generic containers with single argument (list, set, simple tuple) are supported')
            type_arg = type_args[0]
            return [self._cast_single_value(value=val, schema=type_arg, name=self.name) for val in value]
        return self._cast_single_value(value=value, schema=self.schema, name=self.name)

    def _cast_single_value(self, value: Any, schema: type, name: str):
        if self.style is not None and isinstance(value, str):
            value = self._apply_style(value)
        return convert_to_type(value=value, field_type=schema, field_name=name)

    async def get(self, incoming_request: IncomingRequest, security_kwargs: dict[str, Any]) -> Any | NoValue:
        try:
            result = self._get(incoming_request)
        except KeyError:
            if self.default is NoValue:
                if self.is_optional:
                    return NoValue
                raise
            result = self.default
        return self._cast_value_to_schema(result)

    @abc.abstractmethod
    def _get(self, incoming_request: IncomingRequest) -> Any:
        pass

    def _apply_style(self, value: str) -> Any:
        if self.style == ParameterStyle.simple:
            assert isinstance(value, str), "Expected string value, got:"
            return value.split(",")
        return value


@dataclass(frozen=True, kw_only=True)
class QueryParameter(SpecParameter):
    location: ParameterLocation = ParameterLocation.query

    def _get(self, incoming_request: IncomingRequest) -> Any:
        return incoming_request.url_query_key_to_value[self.name]


@dataclass(frozen=True, kw_only=True)
class HeaderParameter(SpecParameter):
    location: ParameterLocation = ParameterLocation.header

    def _get(self, incoming_request: IncomingRequest) -> Any:
        return incoming_request.metadata.header_name_to_value[self.name]


@dataclass(frozen=True, kw_only=True)
class PathParameter(SpecParameter):
    is_optional: bool = field(init=False, default=False)
    location: ParameterLocation = ParameterLocation.path

    def _get(self, incoming_request: IncomingRequest) -> Any:
        return incoming_request.path_match_key_to_value[self.name]


@dataclass(frozen=True, kw_only=True)
class BodyParameter(SpecParameter, abc.ABC):
    location: ParameterLocation = ParameterLocation.body
    body_mime_type: ClassVar[ContentType] = None


@dataclass(frozen=True, kw_only=True)
class JsonWholeBodyParameter(BodyParameter, abc.ABC):
    body_mime_type: ClassVar[ContentType] = ContentType.Json

    def __init__(self):
        raise NotImplementedError("requires inheritance restructuration")


@dataclass(frozen=True, kw_only=True)
class JsonBodyParameter(BodyParameter):
    body_mime_type: ClassVar[ContentType] = ContentType.Json

    def _get(self, incoming_request: IncomingRequest) -> Any:
        raise_if(incoming_request.parsed_body is None, KeyError(self.name))
        return incoming_request.parsed_body[self.name]


@dataclass(frozen=True)
class ParsedBodyPart:
    filename: str | None
    header_name_to_value: dict[str, Any]
    payload: Any
    encoding: str | None
    content_type: ContentType | None


class MultipartParameterHeaders:
    def __init__(self, parent: 'MultipartParameter') -> None:
        self.parent = parent
        self._parent_header_call = CallTemplate(operator.attrgetter("header_name_to_value"), self.parent)

    def __getitem__(self, item: str):
        raise_if(not isinstance(item, str), TypeError(f'Header names must be string, got: {type(item)}: {item}'))
        return CallTemplate(operator.itemgetter(item), self._parent_header_call)

    def __str__(self):
        return f'MultipartParameterHeaders(parent={self.parent})'

    __repr__ = __str__


@dataclass(frozen=True, kw_only=True)
class MultipartParameter(BodyParameter):
    schema: type | UnionType
    body_mime_type: ClassVar[ContentType] = ContentType.MultipartForm
    allowed_content_type: ContentType | tuple[ContentType, ...] | None = None
    headers: tuple[HeaderParameter, ...] = field(init=False)
    unpack: bool = True
    explode: bool | None = None
    raw: bool = False

    def __init__(
            self,
            name: str,
            schema: type | UnionType = bytes,
            is_optional: bool = False,
            default: Any = NoValue,
            description: str | NoValue = NoValue,
            examples: tuple[BaseExample, ...] | NoValue = NoValue,
            allowed_content_type: ContentType | Iterable[ContentType] | None = None,
            headers: Iterable[HeaderParameter] | None = None,
            unpack: bool = True,
            explode: bool | None = None,
            is_filename_required: bool = False,
            style: ParameterStyle | None = None,
            raw: False = False,
    ) -> None:
        if not unpack and default is not NoValue:
            raise_if(not isinstance(default, BodyPart),
                     TypeError(f'With unpack=False default MUST be BodyPart, got {default=}'))
        if explode and typing.get_origin(schema) not in (list, tuple, set):
            raise TypeError(f'explode=True can be used only with iterable schema, got {schema=}')
        object.__setattr__(self, 'name', name)
        object.__setattr__(self, 'schema', schema)
        object.__setattr__(self, 'is_optional', is_optional)
        object.__setattr__(self, 'default', default)
        object.__setattr__(self, 'description', description)
        object.__setattr__(self, 'examples', examples)
        object.__setattr__(self, 'raw', raw)

        if allowed_content_type:
            if is_iterable(allowed_content_type):
                # noinspection PyUnresolvedReferences
                types_tuple = tuple(t.value for t in allowed_content_type)
                object.__setattr__(self, 'allowed_content_type', types_tuple)
                object.__setattr__(
                    self, '_check_content_type',
                    lambda x: raise_if(x not in types_tuple, KeyError(f'Expected {types_tuple}, got: {x!r}')))
            else:
                object.__setattr__(self, 'allowed_content_type', allowed_content_type)
                object.__setattr__(
                    self, '_check_content_type',
                    lambda x: raise_if(x != allowed_content_type, KeyError(f'Expected {allowed_content_type}, got: {x}')))
        else:
            object.__setattr__(self, 'allowed_content_type', None)

        if headers:
            object.__setattr__(self, 'headers', headers_tuple := tuple(headers))
            raise_if(
                len({h.name for h in headers_tuple}) < len(headers_tuple),
                KeyError('Got header parameters with ')
            )
            # object.__setattr__(
            #     self, '_check_headers',
            #     lambda x: raise_if(x not in content_type, KeyError(f'Expected {content_type}, got: {x}')))
        else:
            object.__setattr__(self, 'headers', tuple())
        object.__setattr__(self, 'unpack', unpack)
        object.__setattr__(self, 'explode', explode)
        object.__setattr__(self, 'is_filename_required', is_filename_required)
        object.__setattr__(self, 'style', style)

    def _check_content_type(self, content_type: str | None) -> None:
        pass

    def _check_headers(self, header_name_to_value: CIMultiDictProxy[str]) -> None:
        pass

    def _get(self, incoming_request: IncomingRequest) -> BodyPart | Any:
        raise_if(incoming_request.parsed_body is None, KeyError(self.name))
        raise_if(not isinstance(incoming_request.parsed_body, CIMultiDictProxy),
                 TypeError('Multipart parameter got non multipart body'))
        body: CIMultiDictProxy = incoming_request.parsed_body
        result = body.getall(self.name)
        for body_part in result:
            self._check_content_type(body_part.content_type)
        if not self.explode:
            if len(result) != 1:
                raise ValueError(f"Multipart got {len(result)} values, expected 1")
            result = result[0]

        return result

    def _extract_headers(self, value: BodyPart) -> dict[str, Any]:
        result: dict[str, Any] = {}
        field_name_to_error: dict[str, Exception] = {}
        for header in self.headers:
            if (val := value.get_all_header_values(header.name, header.default)) == NoValue and header.is_required:
                field_name_to_error[header.name] = KeyError("Missing multipart header")
                continue
            if not header.explode:
                if len(val) != 1:
                    field_name_to_error[header.name] = ValueError(f"Multipart header got {len(val)} values, expected 1")
                    continue
                val = val[0]
            result[header.name] = header._cast_value_to_schema(val)

        if field_name_to_error:
            raise FieldErrors(field_name_to_error)
        return result

    def _cast_single_value(self, value: BodyPart, schema: type, name: str) -> ParsedBodyPart:
        if self.unpack:
            if isinstance(value, BodyPart):
                value = value.content if self.raw else value.parse()
            return super()._cast_single_value(value=value, schema=schema, name=name)
        if not isinstance(value, BodyPart):
            raise TypeError(f'UNEXPECTED: Non unpack cast got non BodyPart {value=}')
        return ParsedBodyPart(
            filename=value.filename,
            header_name_to_value=self._extract_headers(value),
            payload=super()._cast_single_value(
                value=value.content if self.raw else value.parse(), schema=schema, name=name),
            encoding=value.encoding,
            content_type=value.content_type
        )

    def _cast_value_to_schema(self, value: Any):
        if self.unpack:
            if isinstance(value, BodyPart):
                return super()._cast_value_to_schema(value=value.parse())
            return super()._cast_value_to_schema(value=value)
        return super()._cast_value_to_schema(value=value)

    @property
    def payload(self) -> 'CallTemplate[bytes]':
        return CallTemplate(operator.attrgetter("payload"), self)

    @property
    def filename(self) -> 'CallTemplate[str | None]':
        return CallTemplate(operator.attrgetter("filename"), self)

    @property
    def content_type(self) -> 'CallTemplate[ContentType | None]':
        return CallTemplate(operator.attrgetter("content_type"), self)

    @property
    def header(self) -> MultipartParameterHeaders:
        return MultipartParameterHeaders(self)


@dataclass(frozen=True, kw_only=True)
class RawBodyParameter(BodyParameter):
    body_mime_type: ClassVar[ContentType | tuple[ContentType, ...]] = ContentType.Octet
    name: str = field(init=False, default=None)
    schema: type = field(init=False, default=bytes)

    def _get(self, incoming_request: IncomingRequest) -> Any:
        return incoming_request.payload

    def __class_getitem__(cls, content_type: ContentType | Iterable[ContentType]) -> type:
        if is_iterable(content_type):
            content_type = tuple(content_type)
        return create_type(prepare_class_name([cls], content_type), [cls], {'body_mime_type': content_type})


class ParameterAggregation(CallParameter, abc.ABC):
    @abc.abstractmethod
    def get_spec_parameters(self) -> set[SpecParameter]:
        pass


@dataclass(frozen=True, kw_only=True)
class CallTemplate(ParameterAggregation):
    func: Callable
    args: tuple[CallParameter, ...]
    kwargs: dict[str, CallParameter]

    def __init__(self, func: Callable, *args: CallParameter, **kwargs: CallParameter):
        object.__setattr__(self, 'func', func)
        object.__setattr__(self, 'args', args)
        object.__setattr__(self, 'kwargs', kwargs)

    async def get(self, incoming_request: IncomingRequest, security_kwargs: dict[str, Any]) -> Any:
        args = await asyncio.gather(*[arg.get(incoming_request, security_kwargs) for arg in self.args])
        kwarg_values = await asyncio.gather(*[v.get(incoming_request, security_kwargs) for v in self.kwargs.values()])
        args = [value for value in args if value is not NoValue]
        kwargs = {key: value for key, value in zip(kwarg_values, self.kwargs) if value is not NoValue}
        result = await acall(self.func(*args, **kwargs))
        return result

    def get_spec_parameters(self) -> set[SpecParameter]:
        result = set()
        for arg in itertools.chain(self.args, self.kwargs.values()):
            result = result.union(arg.get_spec_parameters())
        return result


@dataclass(frozen=True, kw_only=True)
class SecurityParameter(CallParameter):
    """Параметр, чьё значение генерируется объектом Security"""
    name: str

    async def get(self, incoming_request: IncomingRequest, security_kwargs: dict[str, Any]) -> Any:
        value = security_kwargs[self.name]
        return convert_to_type(self.schema, self.name, value)

    def get_spec_parameters(self) -> set[SpecParameter]:
        return set()
